/******************** DECLARACIÓN DE LIBRERIAS*******************************/
#include "ESP8266WiFi.h" //Libreria para la tarjeta de desarrollo nodeMCU
#include "WiFiClient.h"
#include "ESP8266WebServer.h"
#include "DNSServer.h" //Libreria para servidor DNS. Utilizada para el portal cautivo.
#include "PubSubClient.h"  //Libreria MQTT para arduino
#include "EEPROM.h" //Libreria para la memoria EEPROM

/******************** DECLARACION DE VARIABLES*******************************/
//Topics
const char* topic_deltaSensor = "SENSOR/Delta";
const char* topic_MaxEnviarSensor = "SENSOR/Max_enviar";
const char* topic_RastreoSensor = "SENSOR/Rastreo";
const char* topic_tempActual = "SENSOR/Temperatura";

//Variables mensaje
long lastMsg = 0;
long lastMsg2 = 0;
char msg[50];

//Asignacion GPIO a los LED
int LED_AP = 5;
int LED_WiFi = 4;
int ONLed = 13;
int pulsador_reset = 12;

//Variable intentos de conexion wifi STA, cada intento corresponde a medio segundo.
int intentos = 0;

//Variable de estado del dispositivo:
//  0-Crear AP
//  1-AP creado, esperando recibir las credenciales
//  2-Credenciales recibidas, se pasa a modo STA y se conecta al WiFi
//  3-Conectado al WiFi, el sensor llevará a cabo sus tareas especificas
int estado=0;

//Contador de pulsos del pulsador
volatile int cont = 0;

//Variable sensor temperatura LM35
int pinLM35 = 0;
float tempC; //Valor de la temperatura obtenida del sensor
float old_tempC=0; //Último valor enviado(publicado)

//Variables recibidas 
long rastreo = 2000;
long max_envio = 10000;
float delta = 1.0;

//Variables para la construccion de la pagina web
String cabecera =
"<!DOCTYPE HTML >"
"<html>"
"  <head>"
"    <title>TempSensor - WiFi Setup</title>"
"    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">"
"    <meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">"
"  </head>"
"  <body style=\"background-color:#F1EFC8\";>"
"    <div>"
"      <div style=\"background-color:#C70039\"; align=\"middle\">"
"        <h2 style=\"color:#F1EFC8\";><br>&#127777;CONTROL<br>DE<br>TEMPERATURA<br></h2>"
"        <h2 style=\"color:#F1EFC8\";>[Sensor de Temperatura]<br><br>WiFi SetUp<br><br></h2>"
"      </div>";

String lista_ssid = "";

String pssw_broker = 
"          <tr>"
"             <td>&#9656;</td>"
"             <th align=\"left\">CONTRASE&Ntilde;A:</th>"
"          </tr>"
"          <tr>"
"             <td></td>"
"             <td><input name=\"password\" type=\"password\" id=\"myInput\"></td>"
"             <td align=\"center\"><input type=\"checkbox\" onclick=\"myFunction()\"></td>"
"             <td align=\"center\"> &#x1f441;</td>"
"          </tr>"
"          <tr>"
"             <td>&#9656;</td>"
"             <th align=\"left\">BROKER:</th>"
"          </tr>"
"          <tr>"
"             <td></td>"
"             <td><input name=\"broker\" type=\"text\"></td>"
"             <td></td>"
"             <td></td>"
"          </tr>"
"    </table><br>";

String botones = 
"            <div align=\"middle\">"
"             <input type=\"submit\" value=\"Enviar\">"
"             <input type=\"reset\" value=\"Borrar\">"
"             <br><br>"
"             <p>*En caso de no aparecer la red WiFi deseada, refresque la pagina refresque la pagina o escanee de nuevo pulsando:[ &#8634; ].</p>"
"            </div>";

String callback_submit =
"      <div  align=\"middle\">"
"        <h3>&#10003; Credenciales enviados.</h3><br><br><br><br><br><br>"
"      </div>";

String pie_pagina =
"           <div style=\"background-color:#FFA630\"; align=\"middle\">"
"             <p><i><br>Configuraci&oacute;n de la conexi&oacute;n<br><br></i></p>"
"           </div>"
"        </form>"
"      </div>"
"    </div>"
"  <script>"
"  function myFunction() {"
"    var x = document.getElementById(\"myInput\");"
"    if (x.type === \"password\") {"
"      x.type = \"text\";"
"    } else {"
"      x.type = \"password\";"
"    }"
"  }"
"  </script>"
"  </body>"
"</html>";

//Datos conexión punto de acceso (AP) WiFi
const char ssid_ap[] = "TempSensor_WiFi_setup";
const char password_ap[] = "0123456789";

//Variables almacenamiento credenciales red WiFi de casa
char ssid[50];
char password[50];

//Datos del broker MQTT
char mqtt_server[50];

//Declaración servidor DNS
const byte DNS_PORT = 53;
DNSServer dnsServer;

//Declaracion servidor
ESP8266WebServer server(80);

//Cliente wifi
WiFiClient espClient;
PubSubClient client(espClient);

/******************** DECLARACIÓN DE FUNCIONES*******************************/

//Función WIFI AP punto de acceso wifi
void setup_wifi_ap(){
  boolean result_ap = false; //Variable resultado de intentar crear access point;
  digitalWrite(LED_WiFi, LOW);
  digitalWrite(LED_AP, HIGH);
  Serial.print("\n\n---> MODO WIFI AP\n\n");

  while(result_ap==false){
    WiFi.disconnect(); //Desconectamos el WiFi
    WiFi.mode(WIFI_AP); //Establecemos el modo de WiFi como Access Point
    result_ap = WiFi.softAP(ssid_ap,password_ap); //Creamos el punto de acceso wifi
    delay(100);
    if(result_ap == true){    
      Serial.println("Access Point creado con exito\n");
      Serial.print("Direccion IP Access Point - por defecto: ");      //Imprime la dirección IP
      Serial.println(WiFi.softAPIP()); 
      Serial.print("Direccion MAC Access Point: ");                   //Imprime la dirección MAC
      Serial.println(WiFi.softAPmacAddress());

      dnsServer.start(DNS_PORT, "*",IPAddress(192, 168, 4, 1)); //Arranco el servidor DNS para el portal cautivo
                                                                //en la dirección IP 192.168.4.1 del NodeMCU
      server.onNotFound([]() {
        escanear_redes();
      });

      //Declaramos las llamadas a las funciones de las peticiones HTTP
      server.on("/", handleRoot);
      server.on("/", escanear_redes);
      server.on("/submit", handleSubmit);
      server.on("/escanear", escanear_redes); //Escanea las redes wifi disponibles
      server.begin();
      Serial.println("Servidor HTTP creado");
      estado = 1; //Se cambia el estado a 1. El AP ha sido creado y se mantiene a la espera de las credenciales.
    }
    else
    {
      Serial.println("Fallo al crear el Access Point");
      Serial.print("\nReintentando");delay(200);Serial.print(".");delay(200);Serial.print(".");delay(200);Serial.println(".");
    }
  }
}//Fin setup_wifi_ap()

//Función de envío de pagina web cuando se esta en AP mode
void handleRoot()
{
  server.send(200, "text/html", cabecera + lista_ssid + pssw_broker + botones + pie_pagina);
}//Fin handleRoot()

//Función de recepción y almacenamiento de SSID y PASSWORD y BROKER de la pag web
void handleSubmit(){ 
  String SSIDvalue;
  String PASSWORDvalue;
  String BROKERvalue;

  Serial.println("\n**DATOS RECIBIDOS PAGINA WEB**\n");
  
  if (server.args() > 0 ) {
    for ( int i = 0; i < server.args(); i++ ) {
      if (server.argName(i) == "ssid") {
         SSIDvalue = server.arg(i);
         SSIDvalue.toCharArray(ssid, 50);
         Serial.print("SSID recibido: ");
         Serial.println(ssid); 
      }else if(server.argName(i) == "password"){
         PASSWORDvalue = server.arg(i);
         PASSWORDvalue.toCharArray(password, 50);
         Serial.print("PASSWORD recibida: ");
         Serial.println(password);
      }else if(server.argName(i) == "broker"){
         BROKERvalue = server.arg(i);
         BROKERvalue.toCharArray(mqtt_server, 50);
         Serial.print("BROKER recibido: ");
         Serial.println(mqtt_server);
      }
    }
  }

  Serial.println("\n******************************\n");

  //Cuando se reciben los datos, se lanza una pagina web que muestra que los datos han sido recibidos
  server.send(200, "text/html", cabecera + callback_submit +  pie_pagina);
  delay(100);
  guardarCredenciales(); //Se guardan los credenciales en la memoria EEPROM
  cont=0; //Ponemos el contador de pulsos del pulsador a cero
  estado=2; //Se cambia a estado 2. Las credenciales han sido recibidas y se pasa al modo STA para conectarse al WiFi
}//Fin handleSubmit()

//Función que escanea las redes WIFI disponibles y genera su parte de código html
void escanear_redes(){
  int n = WiFi.scanNetworks(); //Número de redes encontradas
  if(n == 0){ //Si no hay redes encontradas muestra un mensaje de que no hay redes disponibles
    Serial.println("No se han encontrado redes WiFi disponibles");
    lista_ssid = "No se han encontrado redes WiFi disponibles";
  }else{ //Si encuentra redes WiFi, compone la parte de codigo html para la pagina web
    lista_ssid =
    "      <div  align=\"middle\">"
    "        <h3>Por favor, complete los credenciales:</h3>" 
    "        <form action=\"http://192.168.4.1/submit\" method=\"post\">"
    "        <table>"
    "        <tr>"
    "           <td>&#9656;</td>"
    "           <th align=\"left\">SSID:</th>"
    "           <td></td>"
    "           <td></td>"
    "        </tr>"
    "        <tr>"
    "           <td></td>"
    "           <td><select name=\"ssid\">";
    
    for(int i = 0; i < n; i++){ //Genera la lista de redes WiFi
      lista_ssid = (lista_ssid) + "<option value= \"" + WiFi.SSID(i) + "\">" + WiFi.SSID(i)+ "</option>";
      delay(10);
    }
    //Se forma la parte de la página web de la lista de redes
    lista_ssid = (lista_ssid) + "</select></td>";
    lista_ssid = (lista_ssid) + "<td align=\"center\"><button type=\"submit\" formaction=\"http://192.168.4.1/escanear\">&#8634;</button></td>";
    lista_ssid = (lista_ssid) + "<td align=\"center\"> *</td></tr>";
    
    handleRoot(); //Llama a la función que genera y envía al navegador la pagina completa
  } 
}//Fin escanear_redes()

//Función WIFI STA conexión a red wifi externa
void setup_wifi(){
  digitalWrite(LED_AP, HIGH);
  bool estado_LED_WiFi = LOW;
  long lastMsg = 0;
  Serial.print("\n\n---> MODO WIFI STA\n\n");

  Serial.print("Conectando a... ");
  Serial.println(ssid);
  
  WiFi.softAPdisconnect(); //Desconectamos el modo AP del wifi
  WiFi.disconnect(); //Desconectamos el wifi
  
  WiFi.mode(WIFI_STA); //Establecemos modo wifi STA
  WiFi.begin(ssid, password); //Conexión a la red WiFi

  //Mientras no se ha conectado al WiFi espera 30 segundos
  while (WiFi.status() != WL_CONNECTED){
    Serial.print(".");
    delay(500);
    intentos ++;
    digitalWrite(LED_WiFi, estado_LED_WiFi);
    estado_LED_WiFi = !estado_LED_WiFi; //Se cambia el estado para que el led parpadee
    //Cada intento equivale a medio segundo: 60/2= 30 segundos
    if(intentos > 60){ //Si supera los 30 segundos, es decir los 60 intentos volverá al modo AP
      Serial.println("\n¡No se pudo conectar!\nEs posible que los credenciales sean invalidos.");
      Serial.println("*NOTA: Entrara en el modo AP de nuevo. Configure la red WIFI nuevamente.");
      estado = 0; //Se cambia estado 0, que generará de nuevo el punto de acceso AP
      intentos = 0; //Pone el contador de tiempo a cero
      return; //Se sale de la función para que no ejecute el codigo de abajo
    }
  }

  //Si se consigue conectar al WiFi se ejecutara el siguiente código:
  
  digitalWrite(LED_WiFi, HIGH); //Se enciende el led de conexión que indica que esta conectado al WiFi
  
  Serial.println("");
  Serial.print("Conectado a la red:");
  Serial.println(ssid);
  Serial.print("Direccion IP: ");
  Serial.println(WiFi.localIP());
  Serial.println("\n");
  estado = 3; // Se cambia al estado 3. Se ha conectado al WiFi y el sensor hará sus tareas.
  intentos = 0;
  
}//Fin setup_wifi()

//Función cargar credenciales de la EEPROM
int cargarCredenciales() {
  EEPROM.begin(512);
  EEPROM.get(0, ssid);
  EEPROM.get(0 + sizeof(ssid), password);
  EEPROM.get(0 + sizeof(ssid) + sizeof(password), mqtt_server);
  char ok[4 + 1]; //Esta variable nos sirve para comprobar si hay credenciales almacenadas
  EEPROM.get(0 + sizeof(ssid) + sizeof(password) + sizeof(mqtt_server), ok);
  EEPROM.end();
  
  //Si no coindice la variable ok con la cadena de caracateres no hay nada almacenado
  if (String(ok) != String("OKEY")) { 
    ssid[0] = 0;
    password[0] = 0;
    mqtt_server[0] = 0;
    return 0; //Devuelve 0 cuando no hay credenciales almacenadas
  }
  
  //Si hay credenciales almacenadas sobreescribe las variables de programa de estas
  Serial.println("Credenciales recuperados:");
  Serial.println(ssid);
  Serial.println(strlen(password) > 0 ? "********" : "<no password>");
  Serial.println(mqtt_server);
  return 2; //Devuelve 2 cuando hay credenciales almacenadas
}//Fin cargarCredenciales()

//Función que guarda los credenciales en la EEPROM
void guardarCredenciales() {
  EEPROM.begin(512);
  EEPROM.put(0, ssid);
  EEPROM.put(0 + sizeof(ssid), password);
  EEPROM.put(0 + sizeof(ssid) + sizeof(password), mqtt_server);
  char ok[4 + 1] = "OKEY"; //Esta variable nos sirve para comprobar si hay credenciales almacenadas
  EEPROM.put(0 + sizeof(ssid) + sizeof(password) + sizeof(mqtt_server), ok);
  EEPROM.commit();
  EEPROM.end();
}//Fin guardarCredenciales()

//Función que elimina los credenciales almacenados en la EEPROM
void eliminarCredenciales() {
  bool estado_LED_AP = LOW;
  EEPROM.begin(512);
  //Para borrar las credenciales sobreescribimos la memoria EEPROM con ceros
  for (int i = 0; i < 512; i++){
    EEPROM.write(i, 0);
  }
  //Parpadea para dar una señal visual de que se ha borrado la EEPROM
  for (int j=0; j<6; j++){
    digitalWrite(LED_AP, estado_LED_AP);
    delay(500);
    estado_LED_AP = !estado_LED_AP;
  }
  Serial.println("Credenciales borrados de la memoria EEPROM");
  EEPROM.end();
}//Fin eliminarCredenciales()

//Función para resetear el NodeMCU (con la interrupcion del pulsador) y volver al estado de creacion punto de acceso AP
void Reset(){
  estado=0; //Se pasa al estado 0, de creación del punto de acceso AP.
}//Fin Reset()

//Función que cuenta las veces que pulsas el pulsador (interrupcion en el modo AP)
void contador(){
  cont++;
}//Fin contador()

//Función que recibe los mensajes de los topic suscritos
void callback(char* topic, byte* payload, unsigned int length) {
  String s; //Auxiliar para la conversión del payload
  
  Serial.print("Mensaje recibido [");
  Serial.print(topic);
  Serial.print("]: ");

  if(strcmp(topic,topic_deltaSensor) == 0){ //con strcmp comparo los dos string
    payload[length] = '\0';
    s = String((char*)payload);
    delta  = s.toFloat();
    Serial.println(delta);
  }else if(strcmp(topic,topic_MaxEnviarSensor) == 0){
    payload[length] = '\0';
    s = String((char*)payload);
    max_envio = s.toInt();
    Serial.println(max_envio);
  }else if(strcmp(topic, topic_RastreoSensor)== 0){
    payload[length] = '\0';
    s = String((char*)payload);
    rastreo = s.toInt();
    Serial.println(rastreo);
  }
}//Fin callback()

//Función en la que el cliente se reconecta en caso de desconectarse o no estar conectado
bool reconnect() {
  digitalWrite(LED_AP, HIGH);
  // Creain ID de cliente aleatorio
  String clientId = "ESP8266Client-";
  clientId += String(random(0xffff), HEX);
  // Bucle hasta que se conecte al broker
  while (!client.connected()) {
    Serial.print("Esperando conexion MQTT...");

    // Attempt to connect
    if (client.connect(clientId.c_str())) {
      digitalWrite(LED_AP, LOW);
      Serial.println("conectado");
      return true;
    } else {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" intentar en 5 segundos");
      // Wait 5 seconds before retrying
      delay(5000);
    }
    
    if(estado == 0){ //Si pulso el pulsador de cambio de modo
                   // salgo del while y del reconnect para poder cambiar al modo AP
      return false;
    }
  }
}//Fin reconnect()

//Función que se suscribe a los topic preconfigurados
void subscripcion_topics(){
  Serial.println("Me suscribo a los TOPIC preconfigurados");

  client.subscribe(topic_deltaSensor); // Diferencia de variacion entre el ultimo valor enviado y la ultima lectura del sensor (temp)
  client.subscribe(topic_MaxEnviarSensor); // Tiempo máximo de envío de la temperatura si no envia la tª por modificacion con respecto al delta
  client.subscribe(topic_RastreoSensor); //Tiempo de obtención de un valor del sensor.
}//Fin subscripcion_topics

//Función lectura temperatura
void leer_temperatura(){
  tempC = analogRead(pinLM35); //Leemos el valor analógico del sensor LM35
  tempC = (tempC*330.0)/1024.0;//Calculamos la temperatura a partir del valor analógico con Vin=3.3V

  // Envia el dato al puerto serial
  Serial.print("Dato leido de temp: ");
  Serial.println(tempC);
}//Fin leer_temperaturas

//Función que compara dos variables tipo float en funcion de epsilon(numero de decimales que compara)
//Devuelve un 1 si el primero es mayor que el segundo
//Devuelve un 2 si el segundo es mayor que el primero
//Devuelve un 0 si son iguales
int compara_mayor(float primero, float segundo, int epsilon){
  int primer_comp, segun_comp,valor;
  
  primer_comp = primero * epsilon;
  segun_comp = segundo * epsilon;

  if(primer_comp>segun_comp){
    valor = 1;
  }else if(primer_comp<segun_comp){
    valor = 2;
  }else{
    valor = 0;
  }
  return valor;
}//Fin compara_mayor()

/*********************************** DECLARACIÓN DEL SETUP() *******************************************/
void setup() {
  delay(1000);
  Serial.begin(115200);
  randomSeed(micros());

  //Define los pines que se van a usar como salidas o entradas
  pinMode(LED_AP, OUTPUT);
  pinMode(ONLed, OUTPUT);
  pinMode(LED_WiFi, OUTPUT);
  pinMode(pulsador_reset, INPUT);

  Serial.println("\t SENSOR Tª ENCENDIDO");
  digitalWrite(ONLed, HIGH);
  Serial.println ("\t Comprobacion EEPROM...");
  estado = cargarCredenciales(); //Comprueba si hay credenciales almacenadas en la EEPROM y asigna el estado del dispositivo
  if(estado == 0){
    Serial.println("EEPROM vacia. Nada que recuperar");
  }
}//Fin setup()

/*********************************** DECLARACION DEL LOOP() *******************************************/
void loop() {
  //En función de la variable estado se determina en que modo se encuentra el dispositivo
  switch(estado){
    //Modo 0 - Modo creación AP: El dispositivo crea el punto de acceso AP
    case 0:
      setup_wifi_ap();
      break;//Fin case 0
      
    //Modo 1 - Modo AP: El punto de acceso está creado, se lanza la pagina web y se espera recibir las credenciales
    case 1:
      attachInterrupt(digitalPinToInterrupt(pulsador_reset),contador,FALLING); //Se declará la interrupción para borrar la EEPROM
      
      //Si el pulsador se presiona 3 veces se borra la EEPROM
      if(cont>=3){
        Serial.println("PULSADOR ACCIONADO 3 VECES!!");
        eliminarCredenciales();
        cont=0;
        detachInterrupt(digitalPinToInterrupt(pulsador_reset)); //Se elimina la interrupción para borrar la EEPROM
      }

      dnsServer.processNextRequest(); //Portal cautivo
      server.handleClient();
      detachInterrupt(digitalPinToInterrupt(pulsador_reset)); //Se elimina la interrupción para borrar la EEPROM
      break;//Fin case 1

    //Modo 2 - Modo STA: Credenciales recibidas, se conecta a la red WiFi introducida por el usuario
    case 2:
      setup_wifi();

      break;//Fin case 2

    //Modo 3 - Modo funcionamiento dispositivo. El sensor medirá y enviará la temperatura
    case 3:
      attachInterrupt(digitalPinToInterrupt(pulsador_reset),Reset,FALLING);//Activa interrupcion que accede al modo configuracion wifi (modo AP)
      client.setServer(mqtt_server, 1883); //Crea el cliente MQTT y le indica a que broker se tiene que conectar y en que puerto
      client.setCallback(callback); //Setea el callback de mensajes recibidos de los topic

      //Si el cliente no se conecta reintenta la conexión
      if(!client.connected()){
        bool aux = reconnect();
        if(!aux){
          detachInterrupt(digitalPinToInterrupt(pulsador_reset));//Al salir del estado 3 desactivo la interrupcion
          break;
        }
        subscripcion_topics();//Me subscribo a los topic preconfigurados si se ha conectado al broker
      }
      client.loop();

      //Leer temperatura
      long now = millis();
      if(now - lastMsg > rastreo){
        leer_temperatura();
        lastMsg = now;
      }

      //Enviar temperatura por tiempo o por incremento de temperatura en valor absoluto
      long now2 = millis();
      float resta = fabs(tempC - old_tempC);

      if((compara_mayor(delta,resta,1000)== 2) || (now2 - lastMsg2 > max_envio)){
        
        Serial.println("\nResta temperaturas: ");
        Serial.print(resta);
        Serial.println("\n");
        old_tempC = tempC;
        int aux=old_tempC*1000000;
       
        snprintf (msg, 75, "%d", aux/1000000);
        Serial.print("\nTemperatura enviada: ");
        Serial.print(msg);
        Serial.println("\n");
        client.publish(topic_tempActual, msg, true);
        lastMsg2 = now2;
      }
      
      if(WiFi.status() != WL_CONNECTED){
        Serial.println("Conexion WIFI perdida...Se reintentara de nuevo...");
        detachInterrupt(digitalPinToInterrupt(pulsador_reset));//Al salir del estado 3 desactivo la interrupcion
        estado = 2;
      }
      break;//Fin case 3
  }
  
}//Fin loop()
